<?php

namespace App\Tests\functional;

use Symfony\Bundle\FrameworkBundle\KernelBrowser;
use Symfony\Bundle\FrameworkBundle\Test\WebTestCase;
use Symfony\Component\DomCrawler\Crawler;
use Symfony\Component\Mime\RawMessage;

class ResetPasswordControllerTest extends WebTestCase
{
    private KernelBrowser $client;
    private const VERIFIED_USER_EMAIL = 'defaultUser@localhost.com';
    private const UNVERIFIED_USER_EMAIL = 'notRegistered@localhost.com';
    private const NEW_PASSWORD = "newpassword";

    protected function setUp(): void
    {
        self::ensureKernelShutdown();
        $this->client = static::createClient();
        $this->client->enableProfiler();
    }

    public function testVerifiedUserShouldReceiveAnEmailLinkToResetHisPassword(): void
    {
        $emailMessage = $this->getEmailMessageWithResetPasswordLinkForVerifiedUser();

        $this->assertEmailCount(1, null);
        $this->assertEmailTextBodyContains($emailMessage, 'To reset your password, please visit the following link');
        $this->assertEmailAddressContains(
            $emailMessage,
            'To',
            self::VERIFIED_USER_EMAIL,
            'When resetting a password, an already registered user should receive an email with the reset password link'
        );
    }

    public function testUnverifiedUserShouldNotReceiveAnEmailLinkToResetHisPassword(): void
    {
        $this->submitResetPasswordFormWithAnUnverifiedEmailAddress();

        $this->assertEmailCount(0);
    }

    public function testUserShouldBeAbleToLoginWithHisNewPasswordAfterResettingHisPassword(): void
    {
        $this->userShouldBeRedirectedToTheResetPasswordPageWhenClickingOnEmailLink();

        $crawler = $this->client->request('GET', '/reset-password/reset');
        $reset_password_submit_button = $crawler->selectButton('Reset password');
        $reset_password_form = $reset_password_submit_button->form();

        $reset_password_form['change_password_form[plainPassword][first]'] = self::NEW_PASSWORD;
        $reset_password_form['change_password_form[plainPassword][second]'] = self::NEW_PASSWORD;
        $this->client->submit($reset_password_form);

        $this->assertResponseRedirects('/login');
        $this->client->followRedirect();
        $this->assertResponseIsSuccessful("/login");

        $this->testRegisteredUserWithEnabledAccountIsRedirectedToCapsulesPageWhenSubmittingLoginForm();
    }

    private function getEmailMessageWithResetPasswordLinkForVerifiedUser(): RawMessage
    {
        $crawler = $this->client->request('GET', '/reset-password');
        $this->assertResponseIsSuccessful();

        $submit_button = $crawler->selectButton('Reset password');
        $form = $submit_button->form();

        $form['reset_password_request_form[email]'] = self::VERIFIED_USER_EMAIL;
        $this->client->submit($form);

        $email_message = $this->getMailerMessage(0);

        if (null === $email_message) {
            throw new \Exception("Email message could not be found");
        }

        return $email_message;
    }

    private function submitResetPasswordFormWithAnUnverifiedEmailAddress(): void
    {
        $crawler = $this->client->request('GET', '/reset-password');
        $this->assertResponseIsSuccessful();

        $submit_button = $crawler->selectButton('Reset password');
        $form = $submit_button->form();

        $form['reset_password_request_form[email]'] = self::UNVERIFIED_USER_EMAIL;
        $this->client->submit($form);
    }

    private function testRegisteredUserWithEnabledAccountIsRedirectedToCapsulesPageWhenSubmittingLoginForm(): void
    {
        $crawler = $this->client->request('GET', '/login');
        $login_submit_button = $crawler->selectButton('Log in');
        $login_form = $login_submit_button->form();

        $login_form['email'] = self::VERIFIED_USER_EMAIL;
        $login_form['password'] = self::NEW_PASSWORD;

        $this->client->submit($login_form);

        $this->assertResponseRedirects(
            '/en/my_capsules',
            302,
            'Once the user is logged in, he should be redirected to its capsules lists'
        );
        $this->client->followRedirect();
        $this->assertResponseIsSuccessful('/en/my_capsules');
    }

    private function userShouldBeRedirectedToTheResetPasswordPageWhenClickingOnEmailLink(): void
    {
        $emailMessage = $this->getEmailMessageWithResetPasswordLinkForVerifiedUser();

        /** @phpstan-ignore-next-line */
        $crawler = new Crawler($emailMessage->getHtmlBody());
        $crawlerLink = $crawler->selectLink('Reset my password')->link();
        $this->client->click($crawlerLink);

        $this->assertResponseRedirects("/reset-password/reset");
        $this->client->followRedirect();
        $this->assertResponseIsSuccessful("/reset-password/reset");
    }
}
